/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

/*
   DATA LIST command.

   Entry Points:
   cmd_data_list()
   cmd_repeating_data() [eventually]
   data_list_init()
   data_list_done()
   data_list_func()
 */

#include <config.h>
#include <assert.h>
#include <ctype.h>
#include "error.h"
#include "misc.h"
#include "som.h"
#include "var.h"
#include "settings.h"
#include "tokens.h"
#include "common.h"
#include "file-handle.h"
#include "dfm.h"

/* Describes how to parse one variable. */
typedef struct dls_var_spec_struct dls_var_spec;
struct dls_var_spec_struct
  {
    dls_var_spec *next;
    variable *v;		/* Associated variable.  Used only in
				   parsing.  Not safe later. */
    char name[9];		/* Free-format: Name of variable. */
    int rec;			/* Fixed-format: Record number (1-based). */
    int fc, lc;			/* Fixed-format: Column numbers in record. */
    fmt_spec input;		/* Input format of this field. */
    int fv;			/* First value in case. */
    int type;			/* 0=numeric, >0=width of alpha field. */
  };

/* Constants for DATA LIST type. */
/* must match table in data-list.c:cmd_data_list() */
#define DLS_FIXED 0
#define DLS_FREE  1
#define DLS_LIST  2

/* DATA LIST private data structure. */
typedef struct
  {
    trns_header h;
    file_handle *handle;	/* Input file, never NULL. */
    int type;			/* A DLS_* constant. */
    variable *end;		/* Variable specified on END subcommand. */
    int nrec;			/* Number of records. */
    dls_var_spec *spec;		/* Variable parsing specifications. */
    int eof;			/* Whether EOF has been reached. */
  }
data_list_pgm;

/* Holds information on parsing the data file. */
static data_list_pgm dls;

/* Last dls_var_spec in the chain.  Used for building the linked-list. */
static dls_var_spec *next;

static int parse_fixed (void);
static int parse_free (void);
static void dump_fixed_table (void);
static void dump_free_table (void);
static void read_from_data_list (void);
static void cancel_data_list (void);
static void destroy_dls (any_trns *);
static int read_one_case (any_trns *, ccase *);

int
cmd_data_list (void)
{
  /* 0=print no table, 1=print table.  (TABLE subcommand.)  */
  int table = -1;

  match_id (DATA);
  match_id (LIST);

  if (vfm_source != &input_program_source
      && vfm_source != &file_type_source)
    discard_variables ();

  dls.handle = default_handle;
  dls.type = -1;
  dls.end = NULL;
  dls.nrec = 0;
  dls.spec = NULL;
  next = NULL;

  while (token != '/')
    {
      if (match_id (FILE))
	{
	  match_tok ('=');
	  dls.handle = fh_parse_file_handle ();
	  if (!dls.handle)
	    return 0;
	}
      else if (match_id (RECORDS))
	{
	  match_tok ('=');
	  match_tok ('(');
	  force_int ();
	  dls.nrec = tokint;
	  get_token ();
	  match_tok (')');
	}
      else if (match_id (END))
	{
	  if (dls.end)
	    return msg (SE, _("The END subcommand may only be specified once."));
	  match_tok ('=');
	  force_id ();
	  dls.end = find_variable (tokstr);
	  if (!dls.end)
	    dls.end = create_variable (tokstr, NUMERIC, 0);
	  get_token ();
	}
      else if (token == ID)
	{
	  /* must match DLS_* constants in var.h */
	  char id[][8] =
	  {"FIXED", "FREE", "LIST", "NOTABLE", "TABLE", ""};
	  char (*p)[8];
	  int index;

	  for (p = id; (*p)[0]; p++)
	    if (id_match (*p, tokstr))
	      break;
	  if (!(*p)[0])
	    return msg (SE, _("Syntax error."));
	  get_token ();

	  index = p - id;
	  if (index < 3)
	    {
	      if (dls.type != -1)
		return msg (SE, _("Only one of FIXED, FREE, or LIST may "
			    "be specified."));
	      dls.type = index;
	    }
	  else
	    table = index - 3;
	}
      else
	return msg (SE, _("Syntax error."));
    }

  default_handle = dls.handle;

  if (dls.type == -1)
    dls.type = DLS_FIXED;

  if (table == -1)
    {
      if (dls.type == DLS_FREE)
	table = 0;
      else
	table = 1;
    }

  if (dls.type == DLS_FIXED)
    {
      if (!parse_fixed ())
	return 0;
      if (table)
	dump_fixed_table ();
    }
  else
    {
      if (!parse_free ())
	return 0;
      if (table)
	dump_free_table ();
    }

  if (vfm_source != NULL)
    {
      data_list_pgm *new_pgm;

      dls.h.proc = read_one_case;
      dls.h.free = destroy_dls;

      new_pgm = xmalloc (sizeof (data_list_pgm));
      memcpy (new_pgm, &dls, sizeof (data_list_pgm));
      add_transformation ((any_trns *) new_pgm);
    }
  else
    vfm_source = &data_list_source;

  return 1;
}

static void
append_var_spec (dls_var_spec * spec)
{
  if (next == 0)
    dls.spec = next = xmalloc (sizeof (dls_var_spec));
  else
    next = next->next = xmalloc (sizeof (dls_var_spec));

#if __CHECKER__
  spec->type = ROUND_UP (spec->type, 8);
#endif

  memcpy (next, spec, sizeof (dls_var_spec));
  next->next = NULL;
}

/* Fixed-format parsing. */

/* Used for chaining together fortran-like format specifiers. */
typedef struct fmt_list
{
  struct fmt_list *next;
  int count;
  fmt_spec f;
  struct fmt_list *down;
}
fmt_list;

/* Used as "local" variables among the fixed-format parsing funcs.  If
   it were guaranteed that Fiasco were going to be compiled by gcc,
   I'd make all these functions a single set of nested functions. */
static struct
  {
    char **name;		/* variable names */
    int nname;			/* number of names */
    int cname;			/* dump_fmt_list(): index of next name to use */

    int recno;			/* index of current record */
    int sc;			/* 1-based column number of starting column for
				   next field to output */

    dls_var_spec spec;		/* next specification to output */
    int fc, lc;			/* first, last column in set of fields specified
				   together */

    int level;			/* level of nesting in fixed_parse_fortran() */
  }
fx;

static int fixed_parse_compatible (void);
static fmt_list *fixed_parse_fortran (void);

static int
parse_fixed (void)
{
  int i;

  fx.recno = 0;
  fx.sc = 1;

  while (token != '.')
    {
      while (match_tok ('/'))
	{
	  fx.recno++;
	  if (token == NUM)
	    {
	      force_int ();
	      if (tokint < fx.recno)
		return msg (SE, _("The record number specified, %ld, is "
			    "before the previous record, %d.  Data "
			    "fields must be listed in order of "
			    "increasing record number."),
			    tokint, fx.recno - 1);
	      fx.recno = tokint;
	      get_token ();
	    }
	  fx.sc = 1;
	}
      fx.spec.rec = fx.recno;

      if (!parse_DATA_LIST_vars (&fx.name, &fx.nname, PV_NONE))
	return 0;

      if (token == NUM)
	{
	  if (!fixed_parse_compatible ())
	    goto fail;
	}
      else if (token == '(')
	{
	  fx.level = 0;
	  fx.cname = 0;
	  if (!fixed_parse_fortran ())
	    goto fail;
	}
      else
	{
	  msg (SE, _("SPSS-like or FORTRAN-like format "
	       "specification expected after variable names."));
	  goto fail;
	}

      for (i = 0; i < fx.nname; i++)
	free (fx.name[i]);
      free (fx.name);
    }
  if (dls.nrec && next->rec > dls.nrec)
    return msg (SE, _("Variables are specified on records that "
		"should not exist according to RECORDS subcommand."));
  else if (!dls.nrec)
    dls.nrec = next->rec;
  if (token != '.')
    return msg (SE, _("End of command expected."));
  return 1;

fail:
  for (i = 0; i < fx.nname; i++)
    free (fx.name[i]);
  free (fx.name);
  return 0;
}

static int
fixed_parse_compatible (void)
{
  int dividend;
  int i;

  force_int ();
  fx.fc = tokint;
  if (fx.fc < 1)
    return msg (SE, _("Column positions for fields must be positive."));
  get_token ();

  convert_negative_to_dash ();
  if (match_tok ('-'))
    {
      force_int ();
      fx.lc = tokint;
      if (fx.lc < 1)
	return msg (SE, _("Column positions for fields must be positive."));
      else if (fx.lc < fx.fc)
	return msg (SE, _("The ending column for a field must be "
		    "greater than the starting column."));
      get_token ();
    }
  else
    fx.lc = fx.fc;

  fx.spec.input.w = fx.lc - fx.fc + 1;
  if (match_tok ('('))
    {
      fmt_desc *fdp;

      if (token == ID)
	{
	  char *cp;

	  fx.spec.input.type = parse_format_specifier_name (&cp, 0);
	  if (fx.spec.input.type == -1)
	    return 0;
	  if (*cp)
	    return msg (SE, _("A format specifier on this line "
			"has extra characters on the end."));
	  get_token ();
	  match_tok (',');
	}
      else
	fx.spec.input.type = FMT_F;

      if (token == NUM)
	{
	  force_int ();
	  if (tokint < 1)
	    return msg (SE, _("The value for number of decimal places "
			"must be at least 1."));
	  fx.spec.input.d = tokint;
	  get_token ();
	}
      else
	fx.spec.input.d = 0;

      fdp = &formats[fx.spec.input.type];
      if (fdp->n_args < 2 && fx.spec.input.d)
	return msg (SE, _("Input format %s doesn't accept decimal "
		    "places."), fdp->name);
      if (fx.spec.input.d > 16)
	fx.spec.input.d = 16;

      force_match (')');
    }
  else
    {
      fx.spec.input.type = FMT_F;
      fx.spec.input.d = 0;
    }

  fx.sc = fx.lc + 1;

  if ((fx.lc - fx.fc + 1) % fx.nname)
    return msg (SE, _("The %d columns %d-%d "
		"can't be evenly divided into %d fields."),
		fx.lc - fx.fc + 1, fx.fc, fx.lc, fx.nname);

  dividend = (fx.lc - fx.fc + 1) / fx.nname;
  fx.spec.input.w = dividend;
  if (!check_input_specifier (&fx.spec.input))
    return 0;

  for (i = 0; i < fx.nname; i++)
    {
      int type;
      variable *v;

      if (fx.spec.input.type == FMT_A || fx.spec.input.type == FMT_AHEX)
	type = ALPHA;
      else
	type = NUMERIC;

      v = create_variable (fx.name[i], type, dividend);
      if (v)
	{
	  convert_fmt_ItoO (&fx.spec.input, &v->print);
	  v->write = v->print;
	}
      else
	{
	  v = find_variable (fx.name[i]);
	  assert (v);
	  if (!vfm_source)
	    return msg (SE, _("%s is a duplicate variable name."), fx.name[i]);
	  if (type != v->type)
	    return msg (SE, _("There is already a variable %s of a "
			"different type."), fx.name[i]);
	  if (type == ALPHA && dividend != v->width)
	    return msg (SE, _("There is already a string variable %s of a "
			"different width."), fx.name[i]);
	}

      fx.spec.v = v;
      fx.spec.fc = fx.fc + dividend * i;
      fx.spec.lc = fx.spec.fc + dividend - 1;
      fx.spec.fv = v->fv;
      fx.spec.type = v->type == NUMERIC ? 0 : v->width;
      append_var_spec (&fx.spec);
    }
  return 1;
}

/* Destroy a format list and, optionally, all its sublists. */
static void
destroy_fmt_list (fmt_list * f, int recurse)
{
  fmt_list *next;

  for (; f; f = next)
    {
      next = f->next;
      if (recurse && f->f.type == FMT_DESCEND)
	destroy_fmt_list (f->down, 1);
      free (f);
    }
}

static int
dump_fmt_list (fmt_list * f)
{
  int i;

  for (; f; f = f->next)
    if (f->f.type == FMT_X)
      fx.sc += f->count;
    else if (f->f.type == FMT_T)
      fx.sc = f->f.w;
    else if (f->f.type == FMT_NEWREC)
      {
	fx.recno += f->count;
	fx.sc = 1;
      }
    else
      for (i = 0; i < f->count; i++)
	if (f->f.type == FMT_DESCEND)
	  {
	    if (!dump_fmt_list (f->down))
	      return 0;
	  }
	else
	  {
	    int type;
	    variable *v;

	    type = (formats[f->f.type].cat & FCAT_STRING) ? ALPHA : NUMERIC;
	    if (fx.cname >= fx.nname)
	      return msg (SE, _("The number of format "
				"specifications exceeds the number of "
				"variable names given."));
	    fx.spec.v = v = create_variable (fx.name[fx.cname++], type, f->f.w);
	    if (!v)
	      return msg (SE, _("%s is a duplicate variable name."), fx.name[i]);
	    fx.spec.input = f->f;
	    convert_fmt_ItoO (&fx.spec.input, &v->print);
	    v->write = v->print;

	    fx.spec.rec = fx.recno;
	    fx.spec.fc = fx.sc;
	    fx.spec.lc = fx.sc + f->f.w - 1;
	    fx.spec.fv = v->fv;
	    fx.spec.type = v->type == NUMERIC ? 0 : v->width;
	    append_var_spec (&fx.spec);

	    fx.sc += f->f.w;
	  }
  return 1;
}

/* Calls itself recursively to parse nested levels of parentheses.
   Returns to its original caller: NULL, to indicate error; non-NULL,
   but nothing useful, to indicate success (it returns a free()'d
   block). */
static fmt_list *
fixed_parse_fortran (void)
{
  fmt_list *head;
  fmt_list *fl = NULL;

  get_token ();			/* skip opening parenthesis */
  while (token != ')')
    {
      if (fl)
	fl = fl->next = xmalloc (sizeof (fmt_list));
      else
	head = fl = xmalloc (sizeof (fmt_list));

      if (token == NUM)
	{
	  if (tokint == NOT_LONG)
	    {
	      msg (SE, _("Integer expected for repeat count."));
	      goto fail;
	    }
	  fl->count = tokint;
	  get_token ();
	}
      else
	fl->count = 1;

      if (token == '(')
	{
	  fl->f.type = FMT_DESCEND;
	  fx.level++;
	  fl->down = fixed_parse_fortran ();
	  fx.level--;
	  if (!fl->down)
	    goto fail;
	}
      else if (match_tok ('/'))
	fl->f.type = FMT_NEWREC;
      else if (!parse_format_specifier (&fl->f, 1)
	       || !check_input_specifier (&fl->f))
	goto fail;

      match_tok (',');
    }
  fl->next = NULL;
  get_token ();

  if (fx.level)
    return head;

  fl->next = NULL;
  dump_fmt_list (head);
  if (fx.cname < fx.nname)
    {
      msg (SE, _("There aren't enough format specifications "
	   "to match the number of variable names given."));
      goto fail;
    }
  destroy_fmt_list (head, 1);
  return head;

fail:
  fl->next = NULL;
  destroy_fmt_list (head, 0);

  return NULL;
}

static void
dump_fixed_table (void)
{
  dls_var_spec *spec;
  som_submission_form f;
  char *buf, *filename;
  int i;

  som_push_workspace ();

  for (i = 0, spec = dls.spec; spec; spec = spec->next)
    i++;
  f.type = SOM_SUB_COLUMNATE_DOWN;
  f.options = SOM_TOPT_NONE;
  f.t = som_create_table (4, i + 1, SOM_CREATE_NONE);
  som_set_text (f.t, 0, 0, SOPT_A_CNTR | SOT_TITLE, _("Variable"));
  som_set_text (f.t, 1, 0, SOPT_A_CNTR | SOT_TITLE, _("Record"));
  som_set_text (f.t, 2, 0, SOPT_A_CNTR | SOT_TITLE, _("Columns"));
  som_set_text (f.t, 3, 0, SOPT_A_CNTR | SOT_TITLE, _("Format"));
  som_box (f.t, SLIN_1, SLIN_1, SLIN_0, SLIN_1, 0, 0, 3, i);
  som_hline (f.t, SLIN_2, 0, 3, 1);
  som_set_headers (f.t, 0, 0, 1, 0);
  for (i = 1, spec = dls.spec; spec; spec = spec->next, i++)
    {
      som_set_text (f.t, 0, i, SOPT_A_LEFT, spec->v->name);
      som_set_text (f.t, 1, i, SOT_PRINTF, "%d", spec->rec);
      som_set_text (f.t, 2, i, SOT_PRINTF, "%3d-%3d",
		    spec->fc, spec->lc);
      som_set_text (f.t, 3, i, SOPT_A_LEFT | SOT_FIX,
		    fmt_to_string (&spec->input));
    }

  filename = fh_handle_name (dls.handle);
  if (filename == NULL)
    filename = "";
  buf = local_alloc (strlen (filename) + INT_DIGITS + 80);
  sprintf (buf, (dls.handle != inline_file
		 ? _("Reading %3d records from file %s.")
		 : _("Reading %3d records from the command file.")),
	   dls.nrec, filename);
  f.title.l = strlen (buf);
  f.title.s = buf;
  som_submit_table (&f);
  fh_handle_name (NULL);
  local_free (buf);

  som_pop_workspace ();
}

/* Free-format parsing. */

static int
parse_free (void)
{
  dls_var_spec spec;
  fmt_spec in, out;
  char **name;
  int nname;
  int i;
  int type;

#if __CHECKER__
  memset (&spec, 0, sizeof (spec));
#endif
  get_token ();
  while (token != '.')
    {
      if (!parse_DATA_LIST_vars (&name, &nname, PV_NONE))
	return 0;
      if (match_tok ('('))
	{
	  if (!parse_format_specifier (&in, 0) || !check_input_specifier (&in))
	    goto fail;
	  force_match (')');
	  convert_fmt_ItoO (&in, &out);
	}
      else
	{
	  match_tok ('*');
	  in.type = FMT_F;
	  in.w = 8;
	  in.d = 0;
	  out = set_format;
	}

      spec.input = in;
      if (in.type == FMT_A || in.type == FMT_AHEX)
	type = ALPHA;
      else
	type = NUMERIC;
      for (i = 0; i < nname; i++)
	{
	  variable *v;

	  spec.v = v = create_variable (name[i], type, in.w);
	  if (!v)
	    return msg (SE, _("%s is a duplicate variable name."), name[i]);
	  v->print = v->write = out;

	  strcpy (spec.name, name[i]);
	  spec.fv = v->fv;
	  spec.type = type == NUMERIC ? 0 : v->width;
	  append_var_spec (&spec);
	}
      for (i = 0; i < nname; i++)
	free (name[i]);
      free (name);
    }
  if (token != '.')
    return msg (SE, _("End of command expected."));
  return 1;

fail:
  for (i = 0; i < nname; i++)
    free (name[i]);
  free (name);
  return 0;
}

static void
dump_free_table (void)
{
  dls_var_spec *spec;
  som_submission_form f;
  char *buf, *filename;
  int i;

  som_push_workspace ();

  for (i = 0, spec = dls.spec; spec; spec = spec->next)
    i++;
  f.type = SOM_SUB_COLUMNATE_DOWN;
  f.options = SOM_TOPT_NONE;
  f.t = som_create_table (2, i + 1, SOM_CREATE_NONE);
  som_set_text (f.t, 0, 0, SOPT_A_CNTR | SOT_TITLE, _("Variable"));
  som_set_text (f.t, 1, 0, SOPT_A_CNTR | SOT_TITLE, _("Format"));
  som_box (f.t, SLIN_1, SLIN_1, SLIN_0, SLIN_1, 0, 0, 1, i);
  som_hline (f.t, SLIN_2, 0, 1, 1);
  som_set_headers (f.t, 0, 0, 1, 0);
  for (i = 1, spec = dls.spec; spec; spec = spec->next, i++)
    {
      som_set_text (f.t, 0, i, SOPT_A_LEFT, spec->v->name);
      som_set_text (f.t, 1, i, SOPT_A_LEFT | SOT_FIX,
		    fmt_to_string (&spec->input));
    }

  filename = fh_handle_name (dls.handle);
  if (filename == NULL)
    filename = "";
  buf = local_alloc (strlen (filename) + INT_DIGITS + 80);
  sprintf (buf, (dls.handle != inline_file
		 ? _("Reading free-form data from file %s.")
		 : _("Reading free-form data from the command file.")),
	   filename);
  f.title.l = strlen (buf);
  f.title.s = buf;
  som_submit_table (&f);
  fh_handle_name (NULL);
  local_free (buf);

  som_pop_workspace ();
}

/* Input procedure. */ 

/* Pointer to relevant parsing data.  Static just to avoid passing it
   around so much. */
static data_list_pgm *dlsp;

/* Extracts a field from the current position in the current record.
   Fields can be unquoted or quoted with single- or double-quote
   characters.  *RET_LEN is set to the field length, *RET_CP is set to
   the field itself.  After parsing the field, sets the current
   position in the record to just past the field.  Returns 0 on
   failure or a 1-based column number indicating the beginning of the
   field on success. */
static int
cut_field (char **ret_cp, int *ret_len)
{
  char *cp, *ep;
  int len;

  cp = dfm_get_record (dlsp->handle, &len);
  if (!cp)
    return 0;

  ep = cp + len;

  /* Skip leading whitespace. */
  while (isspace (*cp) && cp < ep)
    cp++;
  if (cp >= ep)
    return 0;

  /* Three types of fields: quoted with ', quoted with ", unquoted. */
  if (*cp == '\'' || *cp == '"')
    {
      int quote = *cp;

      *ret_cp = ++cp;
      while (cp < ep && *cp != quote)
	cp++;
      *ret_len = cp - *ret_cp;
      if (cp < ep)
	cp++;
      else
	msg (SW, _("Scope of string exceeds line."));
    }
  else
    {
      *ret_cp = cp++;
      while (cp < ep && !isspace (*cp) && *cp != ',')
	cp++;
      *ret_len = cp - *ret_cp;
    }

  {
    int beginning_column;
    
    dfm_set_record (dlsp->handle, *ret_cp);
    beginning_column = dfm_get_cur_col (dlsp->handle) + 1;
    
    dfm_set_record (dlsp->handle, cp);
    
    return 1;
  }
}

static void
parse_field (dls_var_spec *var_spec, ccase *c,
	     char *field, int len, int column)
{
  /* Field temporary buffer when necessary. */
  char field_temp[256];

  /* Points into field or field_temp[]. */
  char *field_ptr;

  if (len >= var_spec->input.w)
    {
      if (len > var_spec->input.w)
	msg (SW, _("In columns %d through %d, the value for variable "
		   "%s was %d characters too long for a field of type %s.  "
		   "The value was truncated."),
	     dfm_get_cur_col (dlsp->handle) + 1,
	     dfm_get_cur_col (dlsp->handle) + len,
	     var_spec->name, len - var_spec->input.w,
	     fmt_to_string (&var_spec->input));
      field_ptr = field;
    }
  else
    {
      field_ptr = field_temp;
      memcpy (field_temp, field, len);
      memset (&field_temp[len], ' ', var_spec->input.w - len);
    }
  parse_string_as_format (field_ptr, var_spec->input.w, &var_spec->input,
			  column, &c->data[var_spec->fv]);
}

static int read_from_data_list_fixed (void);
static int read_from_data_list_free (void);
static int read_from_data_list_list (void);
static int do_reading (int flag);

/* FLAG==0: reads any number of cases into temp_case and calls
   write_case() for each one, returns garbage.  FLAG!=0: reads one
   case into temp_case and returns -2 on eof, -1 otherwise.
   Uses dlsp as the relevant parsing description. */
static int
do_reading (int flag)
{
  int (*func) (void);

  int code;

  if (dlsp->handle != inline_file)
    dfm_push_cust (dlsp->handle);

  switch (dlsp->type)
    {
    case DLS_FIXED:
      func = read_from_data_list_fixed;
      break;
    case DLS_FREE:
      func = read_from_data_list_free;
      break;
    case DLS_LIST:
      func = read_from_data_list_list;
      break;
    default:
      assert (0);
    }
  if (flag)
    code = func ();
  else
    {
      while (func () != -2)
	if (!write_case ())
	  {
	    debug_printf ((_("abort in write_case()\n")));
	    break;
	  }
      fh_close_handle (dlsp->handle);
#if __CHECKER__
      code = 0;			/* prevent error at `return code;' */
#endif
    }
  if (dlsp->handle != inline_file)
    pop_cust ();

  return code;
}

static int
read_from_data_list_fixed (void)
{
  dls_var_spec *var_spec = dlsp->spec;
  char *line;
  int len;
  int i;

  if (!dfm_get_record (dlsp->handle, NULL))
    return -2;
  for (i = 1; i <= dlsp->nrec; i++)
    {
      line = dfm_get_record (dlsp->handle, &len);
      if (!line)
	{
	  /* Note that this can't occur on the first record. */
	  msg (SW, _("Partial case of %d of %d records discarded."),
	       i - 1, dlsp->nrec);
	  return -2;
	}

      for (; var_spec && i == var_spec->rec; var_spec = var_spec->next)
	parse_string_as_format (&line[var_spec->fc - 1], len - var_spec->fc + 1,
				&var_spec->input, var_spec->fc,
				&temp_case->data[var_spec->fv]);

      dfm_fwd_record (dlsp->handle);
    }

  return -1;
}

static int
read_from_data_list_free (void)
{
  dls_var_spec *var_spec;
  char *field;
  int len;

  for (var_spec = dlsp->spec; var_spec; var_spec = var_spec->next)
    {
      int column;
      
      /* Cut out a field and read in a new record if necessary. */
      for (;;)
	{
	  column = cut_field (&field, &len);
	  if (column != 0)
	    break;

	  if (dfm_get_record (dlsp->handle, NULL))
	    dfm_fwd_record (dlsp->handle);
	  if (!dfm_get_record (dlsp->handle, NULL))
	    {
	      if (var_spec != dlsp->spec)
		msg (SW, _("Partial case discarded.  The first variable "
		     "missing was %s."), var_spec->name);
	      return -2;
	    }
	}
      parse_field (var_spec, temp_case, field, len, column);
    }
  return -1;
}

static int
read_from_data_list_list (void)
{
  dls_var_spec *var_spec;
  char *field;
  int len;

  if (!dfm_get_record (dlsp->handle, NULL))
    return -2;

  for (var_spec = dlsp->spec; var_spec; var_spec = var_spec->next)
    {
      /* Cut out a field and check for end-of-line. */
      int column = cut_field (&field, &len);
      
      if (column == 0)
	{
	  if (set_undefined)
	    msg (SW, _("Missing value(s) for all variables from %s onward.  "
		 "These will be filled with the system-missing value "
		 "or blanks, as appropriate."),
		 var_spec->name);
	  for (; var_spec; var_spec = var_spec->next)
	    if (!var_spec->type)
	      temp_case->data[var_spec->fv].f = SYSMIS;
	    else
	      /* PORTME: See TODO, portability note 1. */
	      memset (temp_case->data[var_spec->fv].s, ' ', var_spec->type);
	  break;
	}
      parse_field (var_spec, temp_case, field, len, column);
    }

  dfm_fwd_record (dlsp->handle);
  return -1;
}

static void
destroy_dls (any_trns * pgm)
{
  data_list_pgm *dls = (data_list_pgm *) pgm;
  dls_var_spec *iter, *next;

  iter = dls->spec;
  while (iter)
    {
      next = iter->next;
      free (iter);
      iter = next;
    }
  fh_close_handle (dls->handle);
}

/* Note that since this is exclusively an input program, C is
   guaranteed to be temp_case. */
static int
read_one_case (any_trns * t, ccase * c)
{
  dlsp = (data_list_pgm *) t;
  return do_reading (1);
}

/* Reads all the records from the data file and passes them to
   write_case(). */
static void
data_list_source_read (void)
{
  dlsp = &dls;
  do_reading (0);
}

/* Destroys the source's internal data. */
static void
data_list_source_destroy_source (void)
{
  destroy_dls ((any_trns *) & dls);
}

case_stream data_list_source = 
  {
    NULL,
    data_list_source_read,
    NULL,
    NULL,
    data_list_source_destroy_source,
    NULL,
    "DATA LIST",
  };
