/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#include <config.h>

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#include <assert.h>
#include <ctype.h>
#include <math.h>
#include "common.h"
#include "error.h"
#include "str.h"
#include "misc.h"
#include "settings.h"
#include "tokens.h"
#include "var.h"

/* Returns ceil(log10(X)), x>0; 1, x==0; undefined, x<0.  Basically
   returns the number of digits in X for nonnegative X. */
int
intlog10 (int x)
{
  long tab[] =
  {10, 100, 1000, 10000, 100000, 1000000, 10000000, 10000000,
   1000000000, 0};
  long *tp;

  int i;

  assert (x >= 0);
  for (i = 1, tp = tab; *tp; i++, tp++)
    if (x < *tp)
      return i;
  assert (0);
#if __GNUC__ || __BORLANDC__
  return 0;
#endif
}

/* Returns a null-terminated string of N spaces. */
char *
spacing (int n)
{
  static char *s;
  static max = 0;

  if (n + 1 > max)
    {
      if (s)
	free (s);
      s = xmalloc (n + 1);
      max = n + 1;
    }
  memset (s, ' ', n);
  s[n] = 0;
  return s;
}

/* Copies at most N-1 characters from SRC to DEST, always appending a
   null character */
void
strmaxcpy (char *dest, char *src, size_t n)
{
  size_t len;

  len = strlen (src);
  if (len <= n - 1)
    strcpy (dest, src);
  else
    /* len>n-1 */
    {
      memcpy (dest, src, n - 1);
      dest[n - 1] = 0;
    }
}

/* Copies SRC to DEST, truncating to N characters or right-padding
   with character PAD to N characters if necessary; does not append a
   null character.  SRC, however, must be null-terminated. */
void
strbarepadcpy (char *dest, char *src, size_t n, int pad)
{
  size_t len;

  len = strlen (src);
  if (len >= n)
    memcpy (dest, src, n);
  else
    {
      memcpy (dest, src, len);
      memset (&dest[len], pad, n - len);
    }
}

/* Copies SRC to DEST, truncating SRC to N characters or right-padding
   with character PAD to N characters if necessary; does not append a
   null character.  SRC must be LEN characters long but does not need
   to be null-terminated. */
void
strbarepadlencpy (char *dest, char *src, size_t n, int pad, size_t len)
{
  if (len >= n)
    memcpy (dest, src, n);
  else
    {
      memcpy (dest, src, len);
      memset (&dest[len], pad, n - len);
    }
}

/* Copies SRC to DEST, truncating SRC to N-1 characters or
   right-padding with character PAD to N-1 characters if necessary;
   always appends a null character */
void
strpadcpy (char *dest, char *src, size_t n, int pad)
{
  size_t len;

  len = strlen (src);
  if (len == n - 1)
    strcpy (dest, src);
  else if (len < n - 1)
    {
      memcpy (dest, src, len);
      memset (&dest[len], pad, n - 1 - len);
      dest[n - 1] = 0;
    }
  else
    /* len>n-1 */
    {
      memcpy (dest, src, n - 1);
      dest[n - 1] = 0;
    }
}

/* Sets the first N-1 characters of S to PAD, and appends a null. */
void
blpstrset (char *s, int n, int pad)
{
  memset (s, pad, n - 1);
  s[n - 1] = 0;
}

fmt_desc formats[FMT_NUMBER_OF_FORMATS + 1] =
{
  /* *INDENT-OFF* */
  {"F",		2,  1,  40,  1,   40, 0001, FMT_F},
  {"N",		2,  1,  40,  1,   40, 0011, FMT_F},
  {"E",		2,  1,  40,  6,   40, 0001, FMT_E},
  {"COMMA",	2,  1,  40,  1,   40, 0001, FMT_COMMA},
  {"DOT",	2,  1,  40,  1,   40, 0001, FMT_DOT},
  {"DOLLAR",	2,  1,  40,  2,   40, 0001, FMT_DOLLAR},
  {"PCT",       2,  1,  40,  2,   40, 0001, FMT_PCT},
  {"Z",		2,  1,  40,  1,   40, 0011, FMT_F},
  {"A",		1,  1, 255,  1,  254, 0004, FMT_A},
  {"AHEX",	1,  2, 254,  2,  510, 0006, FMT_A},
  {"IB",	2,  1,   8,  1,    8, 0010, FMT_F},
  {"P",		2,  1,  16,  1,   16, 0011, FMT_F},
  {"PIB",	2,  1,   8,  1,    8, 0010, FMT_F},
  {"PIBHEX",	2,  2,  16,  2,   16, 0002, FMT_F},
  {"PK",	2,  1,  16,  1,   16, 0010, FMT_F},
  {"RB",	1,  2,   8,  2,    8, 0002, FMT_F},
  {"RBHEX",	1,  4,  16,  4,   16, 0002, FMT_F},
  {"CCA",	2, -1,  -1,  1,   40, 0020, FMT_CCA},
  {"CCB",	2, -1,  -1,  1,   40, 0020, FMT_CCB},
  {"CCC",	2, -1,  -1,  1,   40, 0020, FMT_CCC},
  {"CCD",	2, -1,  -1,  1,   40, 0020, FMT_CCD},
  {"CCE",	2, -1,  -1,  1,   40, 0020, FMT_CCE},
  {"DATE",	1,  9,  40,  9,   40, 0001, FMT_DATE},
  {"EDATE",     1,  8,  40,  8,   40, 0001, FMT_EDATE},
  {"SDATE",     1,  8,  40,  8,   40, 0001, FMT_SDATE},
  {"ADATE",	1,  8,  40,  8,   40, 0001, FMT_ADATE},
  {"JDATE",	1,  5,  40,  5,   40, 0001, FMT_JDATE},
  {"QYR",	1,  4,  40,  6,   40, 0001, FMT_QYR},
  {"MOYR",	1,  6,  40,  6,   40, 0001, FMT_MOYR},
  {"WKYR",	1,  6,  40,  8,   40, 0001, FMT_WKYR},
  {"DATETIME",	2, 17,  40, 17,   40, 0001, FMT_DATETIME},
  {"TIME",	2,  5,  40,  5,   40, 0001, FMT_TIME},
  {"DTIME",	2, 11,  40,  8,   40, 0001, FMT_DTIME},
  {"WKDAY",	1,  2,  40,  2,   40, 0001, FMT_WKDAY},
  {"MONTH",	1,  3,  40,  3,   40, 0001, FMT_MONTH},
  {"T",		1,  1,99999, 1,99999, 0000, FMT_T},
  {"X",         1,  1,99999, 1,99999, 0000, FMT_X},
  {"",         -1, -1,  -1, -1,   -1, 0000, -1},
  /* *INDENT-ON* */
};

int
parse_format_specifier_name (char **cp, int allow_xt)
{
  fmt_desc *f;
  char *ep;
  int x;

  ep = toklongstr;
  while (isalpha (*ep))
    ep++;
  x = *ep;
  *ep = 0;

  for (f = formats; f->name[0]; f++)
    if (streq (f->name, toklongstr))
      {
	int indx = f - formats;

	*ep = x;
	if (cp)
	  *cp = ep;

	if (!allow_xt && (indx == FMT_T || indx == FMT_X))
	  {
	    msg (SE, _("X and T format specifiers not allowed here."));
	    return -1;
	  }
	return indx;
      }

  msg (SE, _("%s is not a valid data format."), toklongstr);
  *ep = x;
  return -1;
}

/* Converts F to its string representation (for instance, "F8.2") and
   returns a pointer to a static buffer containing that string. */
char *
fmt_to_string (fmt_spec * f)
{
  static char buf[32];

  if (formats[f->type].n_args >= 2)
    sprintf (buf, "%s%d.%d", formats[f->type].name, f->w, f->d);
  else
    sprintf (buf, "%s%d", formats[f->type].name, f->w);
  return buf;
}

int
check_input_specifier (fmt_spec * spec)
{
  fmt_desc *f;
  char *str;

  f = &formats[spec->type];
  str = fmt_to_string (spec);
  if (spec->type == FMT_X)
    return 1;
  if (f->cat & FCAT_OUTPUT_ONLY)
    return msg (SE, _("Format %s may not be used as an input format."),
		f->name);
  if (spec->w < f->Imin_w || spec->w > f->Imax_w)
    return msg (SE, _("Input format %s specifies a bad width %d.  "
		"Format %s requires a width between %d and %d."),
		str, spec->w, f->name, f->Imin_w, f->Imax_w);
  if ((f->cat & FCAT_EVEN_WIDTH) && spec->w % 2)
    return msg (SE, _("Input format %s specifies an odd width %d, but "
		"format %s requires an even width between %d and "
		"%d."), str, spec->w, f->name, f->Imin_w, f->Imax_w);
  if (f->n_args > 1 && (spec->d < 0 || spec->d > 16))
    return msg (SE, _("Input format %s specifies a bad number of "
		"implied decimal places %d.  Input format %s allows "
		"up to 16 implied decimal places."), str, spec->d, f->name);
  return 1;
}

int
check_output_specifier (fmt_spec * spec)
{
  fmt_desc *f;
  char *str;

  f = &formats[spec->type];
  str = fmt_to_string (spec);
  if (spec->type == FMT_X)
    return 1;
  if (spec->w < f->Omin_w || spec->w > f->Omax_w)
    return msg (SE, _("Output format %s specifies a bad width %d.  "
		"Format %s requires a width between %d and %d."),
		str, spec->w, f->name, f->Omin_w, f->Omax_w);
  if (spec->d > 1
      && (spec->type == FMT_F || spec->type == FMT_COMMA
	  || spec->type == FMT_DOLLAR)
      && spec->w < f->Omin_w + 1 + spec->d)
    return msg (SE, _("Output format %s requires minimum width %d to allow "
		"%d decimal places.  Try %s%d.%d instead of %s."),
		f->name, f->Omin_w + 1 + spec->d, spec->d, f->name,
		f->Omin_w + 1 + spec->d, spec->d, str);
  if ((f->cat & FCAT_EVEN_WIDTH) && spec->w % 2)
    return msg (SE, _("Output format %s specifies an odd width %d, but "
		"output format %s requires an even width between %d and "
		"%d."), str, spec->w, f->name, f->Omin_w, f->Omax_w);
  if (f->n_args > 1 && (spec->d < 0 || spec->d > 16))
    return msg (SE, _("Output format %s specifies a bad number of "
		"implied decimal places %d.  Output format %s allows "
		"a number of implied decimal places between 1 "
		"and 16."), str, spec->d, f->name);
  return 1;
}

/* If a string variable has width W, you can't display it with a
   format specifier with a required width MIN_LEN>W. */
int
check_string_specifier (fmt_spec * f, int min_len)
{
  if ((f->type == FMT_A && min_len > f->w)
      || (f->type == FMT_A && min_len * 2 > f->w))
    return msg (SE, _("Can't display a string variable of width %d with "
		"format specifier %s."), min_len, fmt_to_string (f));
  return 1;
}

void
convert_fmt_ItoO (fmt_spec * input, fmt_spec * output)
{
  output->type = formats[input->type].output;
  output->w = input->w;
  if (output->w > formats[output->type].Omax_w)
    output->w = formats[output->type].Omax_w;
  output->d = input->d;

  switch (input->type)
    {
    case FMT_F:
    case FMT_N:
      if (output->d > 1 && output->w < 2 + output->d)
	output->w = 2 + output->d;
      break;
    case FMT_E:
      output->w = max (max (input->w, input->d+7), 10);
      output->d = max (input->d, 3);
      break;
    case FMT_COMMA:
    case FMT_DOT:
      /* nothing is necessary */
      break;
    case FMT_DOLLAR:
    case FMT_PCT:
      if (output->w < 2)
	output->w = 2;
      break;
    case FMT_PIBHEX:
      {
	int map[] =
	{4, 6, 9, 11, 14, 16, 18, 21};
	assert (input->w % 2 == 0 && input->w >= 2 && input->w <= 16);
	output->w = map[input->w / 2 - 1];
	break;
      }
    case FMT_RBHEX:
      output->w = 8, output->d = 2;	/* FIXME */
      break;
    case FMT_IB:
    case FMT_PIB:
    case FMT_P:
    case FMT_PK:
    case FMT_RB:
      if (input->d < 1)
	output->w = 8, output->d = 2;
      else
	output->w = 9 + input->d;
      break;
    case FMT_CCA:
    case FMT_CCB:
    case FMT_CCC:
    case FMT_CCD:
    case FMT_CCE:
      assert (0);
    case FMT_Z:
    case FMT_A:
      /* nothing is necessary */
      break;
    case FMT_AHEX:
      output->w = input->w / 2;
      break;
    case FMT_DATE:
    case FMT_EDATE:
    case FMT_SDATE:
    case FMT_ADATE:
    case FMT_JDATE:
      /* nothing is necessary */
      break;
    case FMT_QYR:
      if (output->w < 6)
	output->w = 6;
      break;
    case FMT_MOYR:
      /* nothing is necessary */
      break;
    case FMT_WKYR:
      if (output->w < 8)
	output->w = 8;
      break;
    case FMT_TIME:
    case FMT_DTIME:
    case FMT_DATETIME:
    case FMT_WKDAY:
    case FMT_MONTH:
      /* nothing is necessary */
      break;
    default:
      assert (0);
    }
}

int
parse_format_specifier (fmt_spec * input, int allow_xt)
{
  fmt_spec spec;
  fmt_desc *f;
  char *cp, *cp2;
  int type, w, d;

  if (token != ID)
    return msg (SE, _("Format specifier expected."));
  type = parse_format_specifier_name (&cp, allow_xt);
  if (type == -1)
    return 0;
  f = &formats[type];

  w = strtol (cp, &cp2, 10);
  if (cp2 == cp && type != FMT_X)
    return msg (SE, _("Data format %s does not specify a width."), toklongstr);

  cp = cp2;
  if (f->n_args > 1 && *cp == '.')
    {
      cp++;
      d = strtol (cp, &cp2, 10);
      cp = cp2;
    }
  else
    d = 0;

  if (*cp)
    return msg (SE, _("Data format %s is not valid."), toklongstr);
  get_token ();

  spec.type = type;
  spec.w = w;
  spec.d = d;
  *input = spec;

  return 1;
}

/* The random number generator here is an implementation in C of
   Knuth's Algorithm 3.2.2B (Randomizing by Shuffling) in _The Art of
   Computer Programming_, Vol. 2. */

#define k 13
static int V[k];
static int Y;

static double X2;

/* Initializes the random number generator.  Should be called once by
   every cmd_*() that uses random numbers.  Note that this includes
   all procedures that use expressions since they may generate random
   numbers. */
void
setup_randomize (void)
{
  static time_t curtime;
  int i;

  if (set_seed == NOT_LONG)
    {
      if (!curtime)
	time (&curtime);
      srand (curtime++);
    }
  else
    srand (set_seed);

  set_seed_used = 1;

  for (i = 0; i < k; i++)
    V[i] = rand ();
  Y = rand ();
  X2 = SYSMIS;
}

/* Standard shuffling procedure for increasing randomness of the ANSI
   C random number generator. Returns a random number R where 0 <= R
   <= RAND_MAX. */
inline int
shuffle (void)
{
  int j = k * Y / RAND_MAX;
  Y = V[j];
  V[j] = rand ();
  return Y;
}

/* Returns a random number R where 0 <= R <= X. */
double 
rand_uniform (double x)
{
  return ((double) shuffle ()) / (((double) RAND_MAX) / x);
}

/* Returns a random number from the distribution with mean 0 and
   standard deviation X.  This uses algorithm P in section 3.4.1C of
   Knuth's _Art of Computer Programming_, Vol 2. */
double 
rand_normal (double x)
{
  double U1, U2;
  double V1, V2;
  double S;
  double X1;

  if (X2 != SYSMIS)
    {
      double t = X2;
      X2 = SYSMIS;
      return t * x;
    }
  do
    {
      U1 = ((double) shuffle ()) / RAND_MAX;
      U2 = ((double) shuffle ()) / RAND_MAX;
      V1 = 2 * U1 - 1;
      V2 = 2 * U2 - 1;
      S = V1 * V1 + V2 * V2;
    }
  while (S >= 1);
  X1 = V1 * sqrt (-2. * log (S) / S);
  X2 = V2 * sqrt (-2. * log (S) / S);
  return X1 * x;
}

/* Returns a random integer R, where 0 <= R < X. */
int
rand_simple (int x)
{
  return shuffle () % x;
}

int
strpadcmp (char *a, int a_len, char *b, int b_len)
{
  if (a_len == b_len)
    return memcmp (a, b, a_len);
  else if (a_len > b_len)
    {
      while (a[a_len - 1] == ' ' && a_len > b_len)
	a_len--;
      if (a_len < b_len)
	return -1;
      else if (a_len > b_len)
	return 1;
      else
	return memcmp (a, b, a_len);
    }
  else
    /* a_len<b_len */
    {
      while (b[b_len - 1] == ' ' && a_len < b_len)
	b_len--;
      if (a_len < b_len)
	return -1;
      else if (a_len > b_len)
	return 1;
      else
	return memcmp (a, b, a_len);
    }
}

/* Reverses the order of NBYTES bytes at address P, thus converting
   between little- and big-endian byte orders.  NBYTES should be even
   (ever heard of a 24-bit integer?)  */
void
memrev (void *p, size_t nbytes)
{
  unsigned char *h = p, *t = &h[nbytes - 1];
  unsigned char temp;

  assert ((nbytes & 1) == 0);
  nbytes /= 2;
  while (nbytes--)
    {
      temp = *h;
      *h++ = *t;
      *t-- = temp;
    }
}

/* Finds the last NEEDLE of length NEEDLE_LEN in a HAYSTACK of length
   HAYSTACK_LEN.  Returns a pointer to the needle found. */
char *
memrmem (char *haystack, size_t haystack_len, char *needle, size_t needle_len)
{
  int i;
  for (i = haystack_len - needle_len; i >= 0; i--)
    if (!memcmp (needle, &haystack[i], needle_len))
      return &haystack[i];
  return 0;
}

/* Compares S0 of length S0L to S1 of length S1L.  The shorter string
   is considered to be padded with spaces to the length of the
   longer. */
int
cmp_str (char *s0, int s0l, char *s1, int s1l)
{
  /* 254 spaces. */
  static char blanks[254] =
  "                                                               "
  "                                                               "
  "                                                               "
  "                                                               "
  "  ";

  int diff = s0l - s1l;
  int r;

  if (diff == 0)
    {
      if (s0l == 0)
	return 0;
      return memcmp (s0, s1, s0l);
    }
  else if (diff > 0)
    {
      /* s0l > s1l */
      if (s1l)
	{
	  r = memcmp (s0, s1, s1l);
	  if (r)
	    return r;
	}
      return memcmp (&s0[s1l], blanks, diff);
    }
  else
    /* diff<0 */
    {
      /* s0l < s1l */
      if (s0l)
	{
	  r = memcmp (s0, s1, s0l);
	  if (r)
	    return r;
	}
      return memcmp (blanks, &s1[s0l], -diff);
    }
}

/* Reads a line from FILE into *LINE which can hold *N bytes, in the
   fashion of getline().  Splices lines terminated with `\', then
   deletes comments introduced by `#' outside of single or double
   quotes.  Trailing whitespace will be deleted.  Increments cust_ln
   as appropriate. */
int
get_config_line (char **line, size_t * n, FILE * file)
{
  char *cp;
  int len = getline (line, n, file);

  cust_ln++;
  if (-1 == len)
    return -1;

  while (len && isspace ((*line)[len - 1]))
    len--;
  (*line)[len] = 0;

  if (len == 0 || '\\' != (*line)[len - 1])
    goto exit;
  (*line)[--len] = 0;
  while (1)
    {
      char *line2 = NULL;
      size_t n2 = 0;
      int len2 = getline (&line2, &n2, file);

      cust_ln++;
      if (-1 == len2)
	return len2;

      while (len2 && isspace (line2[len2 - 1]))
	len2--;
      line2[len2] = 0;

      *line = xrealloc (*line, len + len2 + 1);
      *n = len + len2 + 1;
      strcpy (&(*line)[len], line2);
      len += len2;
      free (line2);

      if (len == 0 || '\\' != (*line)[len - 1])
	goto exit;
      (*line)[--len] = 0;
    }

exit:
  cp = strchr (*line, '#');
  if (cp)
    {
      int quote = 0;
      for (cp = *line; *cp; cp++)
	if (quote)
	  {
	    if (*cp == quote)
	      quote = 0;
	    else if (*cp == '\\')
	      cp++;
	  }
	else
	  {
	    if (*cp == '\'' || *cp == '"')
	      quote = *cp;
	    else if (*cp == '#')
	      {
		len = cp - *line;
		while (len && isspace ((*line)[len]))
		  len--;
		(*line)[len] = 0;
		break;
	      }
	  }
    }
  return len;
}

#if 0
/* This function is not currently used, however it might be useful
   someday.  It reverses the order of the COUNT elements of ARRAY,
   each of which is SIZE bytes in size. */
static void
reverse (void *_array, int count, int size)
{
  char *array = _array;
  void *temp;
  int i;

  temp = local_alloc (size);

  for (i = 0; i < count / 2; i++)
    {
      memcpy (temp, &array[i * size], size);
      memcpy (&array[i * size], &array[(count - i) * size], size);
      memcpy (&array[(count - i) * size], temp, size);
    }
  local_free (temp);
}
#endif
